/*
 * Callback.java
 *
 * Copyright (C) 2002-2005 Peter Graves
 * $Id: Function.java 12079 2009-07-31 19:45:54Z ehuelsmann $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent
 * modules, and to copy and distribute the resulting executable under
 * terms of your choice, provided that you also meet, for each linked
 * independent module, the terms and conditions of the license of that
 * module.  An independent module is a module which is not derived from
 * or based on this library.  If you modify this library, you may extend
 * this exception to your version of the library, but you are not
 * obligated to do so.  If you do not wish to do so, delete this
 * exception statement from your version.
 */

package org.armedbear.lisp;

import java.util.concurrent.Callable;

public abstract class Callback extends Function {

    public Callback() {
	super();
    }

    @Override
    public LispObject execute() throws ConditionThrowable {
	try {
	    return JavaObject.getInstance(call());
	} catch(Throwable e) {
	    throw new ConditionThrowable(new JavaException(e));
	}
    }

    protected Object call() throws Throwable {
        return error(new WrongNumberOfArgumentsException(this));
    }

    @Override
    public LispObject execute(LispObject arg0) throws ConditionThrowable {
	try {
	    return JavaObject.getInstance(call(arg0.javaInstance()));
	} catch(Exception e) {
	    throw new ConditionThrowable(new JavaException(e));
	}
    }

    protected Object call(Object arg0) throws Exception, ConditionThrowable {
        return error(new WrongNumberOfArgumentsException(this));
    }

    @Override
    public LispObject execute(LispObject arg0, LispObject arg1) throws ConditionThrowable {
	try {
	    return JavaObject.getInstance(call(arg0.javaInstance(), arg1.javaInstance()));
	} catch(Exception e) {
	    throw new ConditionThrowable(new JavaException(e));
	}
    }

    protected Object call(Object arg0, Object arg1) throws Exception, ConditionThrowable {
        return error(new WrongNumberOfArgumentsException(this));
    }

    /** TODO **/

    public static Callback fromRunnable(final Runnable r) {
	return new Callback() {
	    protected Object call() {
		r.run();
		return null;
	    }
	};
    }

    public static Callback fromCallable(final Callable<?> c) {
	return new Callback() {
	    protected Object call() throws Exception {
		return c.call();
	    }
	};
    }

}